from .ReadHelper import read_int_16le, read_int_8, signed8, signed16


# Do you even embroider .bro?

def read(f, out, settings=None):
    f.seek(0x100, 0)
    count = 0
    while True:
        count += 1
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] != 0x80:
            out.stitch(signed8(b[0]), -signed8(b[1]))
            continue
        control = read_int_8(f)
        if control == 0x00:
            continue
        elif control == 0x02:
            break
        if control != 0x03:
            needle = control - 0xE0
            out.color_change()
        x = signed16(read_int_16le(f))
        y = signed16(read_int_16le(f))
        out.move(x, -y)
    out.end()



import os.path

from .EmbPattern import EmbPattern
import pyembroidery.DstWriter as DstWriter
import pyembroidery.PecWriter as PecWriter
import pyembroidery.PesWriter as PesWriter
import pyembroidery.ExpWriter as ExpWriter
import pyembroidery.Vp3Writer as Vp3Writer
import pyembroidery.JefWriter as JefWriter
import pyembroidery.SvgWriter as SvgWriter
import pyembroidery.CsvWriter as CsvWriter

import pyembroidery.DstReader as DstReader
import pyembroidery.PecReader as PecReader
import pyembroidery.PesReader as PesReader
import pyembroidery.ExpReader as ExpReader
import pyembroidery.Vp3Reader as Vp3Reader
import pyembroidery.JefReader as JefReader

import pyembroidery.XxxReader as XxxReader
import pyembroidery.SewReader as SewReader
import pyembroidery.U01Reader as U01Reader
import pyembroidery.ShvReader as ShvReader
import pyembroidery.A10oReader as A10oReader
import pyembroidery.A100Reader as A100Reader
import pyembroidery.BroReader as BroReader
import pyembroidery.DsbReader as DsbReader
import pyembroidery.DszReader as DszReader
import pyembroidery.EmdReader as EmdReader
import pyembroidery.InbReader as InbReader
import pyembroidery.TbfReader as TbfReader
import pyembroidery.KsmReader as KsmReader
import pyembroidery.TapReader as TapReader
import pyembroidery.StxReader as StxReader


# import pyembroidery.ExyReader as ExyReader
# import pyembroidery.FxyReader as FxyReader
# import pyembroidery.GtReader as GtReader
# import pyembroidery.DatReader as DatReader


def supported_formats():
    """Generates dictionary entries for supported formats. Each entry will
    always have description, extension, mimetype, and category. Reader
    will provide the reader, if one exists, writer will provide the writer,
    if one exists.

    Metadata gives a list of metadata read and/or written by that type.

    Options provides accepted options by the format and their accepted values.
    """
    yield ({
        "description": "Brother Embroidery Format",
        "extension": "pec",
        "mimetype": "application/x-pec",
        "category": "embroidery",
        "reader": PecReader,
        "writer": PecWriter,
        "metadata": ("name")
    })
    yield ({
        "description": "Brother Embroidery Format",
        "extension": "pes",
        "mimetype": "application/x-pes",
        "category": "embroidery",
        "reader": PesReader,
        "writer": PesWriter,
        "options": {
            "pes version": (1, 6),
            "truncated": (True, False)
        },
        "metadata": ("name", "author", "category", "keywords", "comments")
    })
    yield ({
        "description": "Melco Embroidery Format",
        "extension": "exp",
        "mimetype": "application/x-exp",
        "category": "embroidery",
        "reader": ExpReader,
        "writer": ExpWriter,
    })
    yield ({
        "description": "Tajima Embroidery Format",
        "extension": "dst",
        "mimetype": "application/x-dst",
        "category": "embroidery",
        "reader": DstReader,
        "writer": DstWriter,
        "options": {
            "extended headers": (True, False)
        },
        "metadata": ("name")
    })
    yield ({
        "description": "Janome Embroidery Format",
        "extension": "jef",
        "mimetype": "application/x-jef",
        "category": "embroidery",
        "reader": JefReader,
        "writer": JefWriter,
    })
    yield ({
        "description": "Pfaff Embroidery Format",
        "extension": "vp3",
        "mimetype": "application/x-vp3",
        "category": "embroidery",
        "reader": Vp3Reader,
        "writer": Vp3Writer,
    })
    yield ({
        "description": "Scalable Vector Graphics",
        "extension": "svg",
        "mimetype": "image/svg+xml",
        "category": "vector",
        "writer": SvgWriter,
    })
    yield ({
        "description": "Comma-separated values",
        "extension": "csv",
        "mimetype": "text/csv",
        "category": "debug",
        "writer": CsvWriter,
        "options": {
            "deltas": (True, False)
        },
    })
    yield ({
        "description": "Singer Embroidery Format",
        "extension": "xxx",
        "mimetype": "application/x-xxx",
        "category": "embroidery",
        "reader": XxxReader
    })
    yield ({
        "description": "Janome Embroidery Format",
        "extension": "sew",
        "mimetype": "application/x-sew",
        "category": "embroidery",
        "reader": SewReader
    })
    yield ({
        "description": "Barudan Embroidery Format",
        "extension": "u01",
        "mimetype": "application/x-u01",
        "category": "embroidery",
        "reader": U01Reader
    })
    yield ({
        "description": "Husqvarna Viking Embroidery Format",
        "extension": "shv",
        "mimetype": "application/x-shv",
        "category": "embroidery",
        "reader": ShvReader
    })
    yield ({
        "description": "Toyota Embroidery Format",
        "extension": "10o",
        "mimetype": "application/x-10o",
        "category": "embroidery",
        "reader": A10oReader
    })
    yield ({
        "description": "Toyota Embroidery Format",
        "extension": "100",
        "mimetype": "application/x-100",
        "category": "embroidery",
        "reader": A100Reader
    })
    yield ({
        "description": "Bits & Volts Embroidery Format",
        "extension": "bro",
        "mimetype": "application/x-Bro",
        "category": "embroidery",
        "reader": BroReader
    })
    # yield ({
    #     "description": "Barudan Embroidery Format",
    #     "extension": "dat",
    #     "mimetype": "application/x-dat",
    #     "category": "embroidery",
    #     "stable": False,
    #     "reader": DatReader
    # })
    yield ({
        "description": "Tajima(Barudan) Embroidery Format",
        "extension": "dsb",
        "mimetype": "application/x-dsb",
        "category": "embroidery",
        "reader": DsbReader
    })
    yield ({
        "description": "ZSK USA Embroidery Format",
        "extension": "dsz",
        "mimetype": "application/x-dsz",
        "category": "embroidery",
        "reader": DszReader
    })
    yield ({
        "description": "Elna Embroidery Format",
        "extension": "emd",
        "mimetype": "application/x-emd",
        "category": "embroidery",
        "reader": EmdReader
    })
    # yield ({
    #     "description": "Eltac Embroidery Format",
    #     "extension": "exy", # e??, e01
    #     "mimetype": "application/x-exy",
    #     "category": "embroidery",
    #     "stable": False,
    #     "reader": ExyReader
    # })
    # yield ({
    #     "description": "Fortron Embroidery Format",
    #     "extension": "fxy", # f??, f01
    #     "mimetype": "application/x-fxy",
    #     "category": "embroidery",
    #     "stable": False,
    #     "reader": FxyReader
    # })
    # yield ({
    #     "description": "Gold Thread Embroidery Format",
    #     "extension": "gt",
    #     "mimetype": "application/x-exy",
    #     "category": "embroidery",
    #     "stable": False,
    #     "reader": GtReader
    # })
    yield ({
        "description": "Inbro Embroidery Format",
        "extension": "inb",
        "mimetype": "application/x-inb",
        "category": "embroidery",
        "reader": InbReader
    })
    yield ({
        "description": "Tajima Embroidery Format",
        "extension": "tbf",
        "mimetype": "application/x-tbf",
        "category": "embroidery",
        "reader": TbfReader
    })
    yield ({
        "description": "Pfaff Embroidery Format",
        "extension": "ksm",
        "mimetype": "application/x-ksm",
        "category": "embroidery",
        "reader": KsmReader
    })
    yield ({
        "description": "Happy Embroidery Format",
        "extension": "tap",
        "mimetype": "application/x-tap",
        "category": "embroidery",
        "reader": TapReader
    })
    yield ({
        "description": "Data Stitch Embroidery Format",
        "extension": "stx",
        "mimetype": "application/x-stx",
        "category": "embroidery",
        "reader": StxReader
    })


def convert(filename_from, filename_to, settings=None):
    pattern = read(filename_from, settings)
    if pattern is None:
        return
    if settings is not None:
        stable = settings.get("stable", True)
        if stable:
            pattern = pattern.get_stable_pattern()
    else:
        pattern = pattern.get_stable_pattern()
    write(pattern, filename_to, settings)


def get_extension_by_filename(filename):
    """extracts he extension from a filename"""
    return os.path.splitext(filename)[1][1:]


def read_embroidery(reader, f, settings=None, pattern=None):
    """Reads fileobject or filename with reader."""
    if pattern is None:
        pattern = EmbPattern()
    if isinstance(f, str):
        try:
            with open(f, "rb") as stream:
                reader.read(stream, pattern, settings)
                stream.close()
        except IOError:
            pass
    else:
        reader.read(f, pattern, settings)
    return pattern


def read_dst(f, settings=None, pattern=None):
    """Reads fileobject as DST file"""
    return read_embroidery(DstReader, f, settings, pattern)


def read_pec(f, settings=None, pattern=None):
    """Reads fileobject as PEC file"""
    return read_embroidery(PecReader, f, settings, pattern)


def read_pes(f, settings=None, pattern=None):
    """Reads fileobject as PES file"""
    return read_embroidery(PesReader, f, settings, pattern)


def read_exp(f, settings=None, pattern=None):
    """Reads fileobject as EXP file"""
    return read_embroidery(ExpReader, f, settings, pattern)


def read_vp3(f, settings=None, pattern=None):
    """Reads fileobject as VP3 file"""
    return read_embroidery(Vp3Reader, f, settings, pattern)


def read_jef(f, settings=None, pattern=None):
    """Reads fileobject as JEF file"""
    return read_embroidery(JefReader, f, settings, pattern)


def read(filename, settings=None, pattern=None):
    """Reads file, assuming type by extension"""
    extension = get_extension_by_filename(filename)
    extension = extension.lower()
    for file_type in supported_formats():
        if file_type['extension'] != extension:
            continue
        reader = file_type.get("reader", None)
        return read_embroidery(reader, filename, settings, pattern)
    return None


def write_embroidery(writer, pattern, stream, settings=None):
    if settings is None:
        settings = {}
    else:
        settings = settings.copy()
    if settings.get("encode", True):
        if not ("max_jump" in settings):
            try:
                settings["max_jump"] = writer.MAX_JUMP_DISTANCE
            except AttributeError:
                pass
        if not ("max_stitch" in settings):
            try:
                settings["max_stitch"] = writer.MAX_STITCH_DISTANCE
            except AttributeError:
                pass
        if not ("full_jump" in settings):
            try:
                settings["full_jump"] = writer.FULL_JUMP
            except AttributeError:
                pass
        if not ("strip_sequins" in settings):
            try:
                settings["strip_sequins"] = writer.STRIP_SEQUINS
            except AttributeError:
                pass
        pattern = pattern.get_normalized_pattern(settings)

    if isinstance(stream, str):
        with open(stream, "wb") as stream:
            writer.write(pattern, stream, settings)
    else:
        writer.write(pattern, stream, settings)


def write_dst(pattern, stream, settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(DstWriter, pattern, stream, settings)


def write_pec(pattern, stream, settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(PecWriter, pattern, stream, settings)


def write_pes(pattern, stream, settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(PesWriter, pattern, stream, settings)


def write_exp(pattern, stream, settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(ExpWriter, pattern, stream, settings)


def write_vp3(pattern, stream, settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(Vp3Writer, pattern, stream, settings)


def write_jef(pattern, stream, settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(JefWriter, pattern, stream, settings)


def write_svg(pattern, stream, settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(SvgWriter, pattern, stream, settings)


def write(pattern, filename, settings=None):
    """Writes file, assuming type by extension"""
    extension = get_extension_by_filename(filename)
    extension = extension.lower()

    for file_type in supported_formats():
        if file_type['extension'] != extension:
            continue
        writer = file_type.get("writer", None)
        if writer is None:
            continue
        write_embroidery(writer, pattern, filename, settings)



import random

from .EmbThread import EmbThread
from .EmbEncoder import Transcoder as Normalizer
from .EmbConstant import *


class EmbPattern:
    def __init__(self):
        self.stitches = []  # type: list
        self.threadlist = []  # type: list
        self.extras = {}
        # filename, name, category, author, keywords, comments, are typical
        self._previousX = 0  # type: float
        self._previousY = 0  # type: float

    def move(self, dx=0, dy=0):
        """Move dx, dy"""
        self.add_stitch_relative(JUMP, dx, dy)

    def move_abs(self, x, y):
        """Move absolute x, y"""
        self.add_stitch_absolute(JUMP, x, y)

    def stitch(self, dx=0, dy=0):
        """Stitch dx, dy"""
        self.add_stitch_relative(STITCH, dx, dy)

    def stop(self, dx=0, dy=0):
        """Stop dx, dy"""
        self.add_stitch_relative(STOP, dx, dy)

    def trim(self, dx=0, dy=0):
        """Trim dx, dy"""
        self.add_stitch_relative(TRIM, dx, dy)

    def color_change(self, dx=0, dy=0):
        """Color Change dx, dy"""
        self.add_stitch_relative(COLOR_CHANGE, dx, dy)

    def sequin_eject(self, dx=0, dy=0):
        """Eject Sequin dx, dy"""
        self.add_stitch_relative(SEQUIN_EJECT, dx, dy)

    def sequin_mode(self, dx=0, dy=0):
        """Eject Sequin dx, dy"""
        self.add_stitch_relative(SEQUIN_MODE, dx, dy)

    def end(self, dx=0, dy=0):
        """End Design dx, dy"""
        self.add_stitch_relative(END, dx, dy)

    def add_thread(self, thread):
        """Adds thread to design.
        Note: this has no effect on stitching and can be done at any point."""
        if isinstance(thread, EmbThread):
            self.threadlist.append(thread)
        elif isinstance(thread, int):
            thread_object = EmbThread()
            thread_object.color = thread
            self.threadlist.append(thread_object)
        elif isinstance(thread, dict):
            thread_object = EmbThread()
            if "name" in thread:
                thread_object.description = thread["name"]
            if "description" in thread:
                thread_object.description = thread["description"]
            if "desc" in thread:
                thread_object.description = thread["desc"]
            if "brand" in thread:
                thread_object.brand = thread["brand"]
            if "manufacturer" in thread:
                thread_object.brand = thread["manufacturer"]
            if "color" in thread:
                thread_object.color = thread["color"]
            if "rgb" in thread:
                thread_object.color = thread["rgb"]
            if "hex" in thread:
                thread_object.set_hex_color(thread["hex"])
            if "id" in thread:
                thread_object.catalog_number = thread["id"]
            if "catalog" in thread:
                thread_object.catalog_number = thread["catalog"]
            self.threadlist.append(thread_object)

    def metadata(self, name, data):
        """Adds select metadata to design.
        Note: this has no effect on stitching and can be done at any point."""
        self.extras[name] = data

    def get_metadata(self, name, default=None):
        return self.extras.get(name, default)

    def extends(self):
        min_x = float('inf')
        min_y = float('inf')
        max_x = -float('inf')
        max_y = -float('inf')

        for stitch in self.stitches:
            if stitch[0] > max_x:
                max_x = stitch[0]
            if stitch[0] < min_x:
                min_x = stitch[0]
            if stitch[1] > max_y:
                max_y = stitch[1]
            if stitch[1] < min_y:
                min_y = stitch[1]
        return min_x, min_y, max_x, max_y

    def count_stitch_commands(self, command):
        count = 0
        for stitch in self.stitches:
            flags = stitch[2]
            if flags == command:
                count += 1
        return count

    def count_color_changes(self):
        return self.count_stitch_commands(COLOR_CHANGE)

    def count_stitches(self):
        return len(self.stitches)

    def count_threads(self):
        return len(self.threadlist)

    @staticmethod
    def get_random_thread():
        thread = EmbThread()
        thread.color = 0xFF000000 | random.randint(0, 0xFFFFFF)
        thread.description = "Random"
        return thread

    def get_thread_or_filler(self, index):
        if len(self.threadlist) <= index:
            return self.get_random_thread()
        else:
            return self.threadlist[index]

    def get_as_stitchblock(self):
        stitchblock = []
        thread = self.get_thread_or_filler(0)
        thread_index = 1
        for stitch in self.stitches:
            flags = stitch[2]
            if flags == STITCH:
                stitchblock.append(stitch)
            else:
                if len(stitchblock) > 0:
                    yield (stitchblock, thread)
                    stitchblock = []
                if flags == COLOR_CHANGE:
                    thread = self.get_thread_or_filler(thread_index)
                    thread_index += 1
        if len(stitchblock) > 0:
            yield (stitchblock, thread)

    def get_as_command_blocks(self):
        last_pos = 0
        last_command = NO_COMMAND
        for pos, stitch in enumerate(self.stitches):
            command = stitch[2]
            if command == last_command or last_command == NO_COMMAND:
                last_command = command
                continue
            last_command = command
            yield self.stitches[last_pos:pos]
            last_pos = pos
        yield self.stitches[last_pos:]

    def get_as_colorblocks(self):
        thread_index = 0
        last_pos = 0
        for pos, stitch in enumerate(self.stitches):
            if stitch[2] != COLOR_CHANGE:
                continue
            thread = self.get_thread_or_filler(thread_index)
            thread_index += 1
            yield (self.stitches[last_pos:pos], thread)
            last_pos = pos
        thread = self.get_thread_or_filler(thread_index)
        yield (self.stitches[last_pos:], thread)

    def get_unique_threadlist(self):
        return set(self.threadlist)

    def get_singleton_threadlist(self):
        singleton = []
        last_thread = None
        for thread in self.threadlist:
            if thread != last_thread:
                singleton.append(thread)
            last_thread = thread
        return singleton

    def move_center_to_origin(self):
        extends = self.extends()
        cx = round((extends[2] - extends[0]) / 2.0)
        cy = round((extends[3] - extends[1]) / 2.0)
        self.translate(-cx, -cy)

    def translate(self, dx, dy):
        for stitch in self.stitches:
            stitch[0] += dx
            stitch[1] += dy

    def fix_color_count(self):
        """Ensure the there are threads for all color blocks."""
        thread_index = 0
        init_color = True
        for stitch in self.stitches:
            data = stitch[2] & COMMAND_MASK
            if data == STITCH or data == SEW_TO or data == NEEDLE_AT:
                if init_color:
                    thread_index += 1
                    init_color = False
            elif data == COLOR_CHANGE or data == COLOR_BREAK:
                init_color = True
        while len(self.threadlist) < thread_index:
            self.add_thread(self.get_thread_or_filler(len(self.threadlist)))

    def add_stitch_absolute(self, cmd, x=0, y=0):
        """Add a command at the absolute location: x, y"""
        self.stitches.append([x, y, cmd])
        self._previousX = x
        self._previousY = y

    def add_stitch_relative(self, cmd, dx=0, dy=0):
        """Add a command relative to the previous location"""
        x = self._previousX + dx
        y = self._previousY + dy
        self.add_stitch_absolute(cmd, x, y)

    def add_command(self, cmd, x=0, y=0):
        """Add a command, without treating parameters as locations
         that require an update"""
        self.stitches.append([x, y, cmd])

    def add_stitchblock(self, stitchblock):
        threadlist = self.threadlist
        block = stitchblock[0]
        thread = stitchblock[1]
        if len(threadlist) == 0 or thread is not threadlist[-1]:
            threadlist.append(thread)
            self.add_stitch_relative(COLOR_BREAK)
        else:
            self.add_stitch_relative(SEQUENCE_BREAK)

        for stitch in block:
            try:
                self.add_stitch_absolute(stitch.command, stitch.x, stitch.y)
            except AttributeError:
                self.add_stitch_absolute(stitch[2], stitch[0], stitch[1])

    def get_stable_pattern(self):
        """Gets a stablized version of the pattern."""
        stable_pattern = EmbPattern()
        for stitchblock in self.get_as_stitchblock():
            stable_pattern.add_stitchblock(stitchblock)
        stable_pattern.extras.update(self.extras)
        return stable_pattern

    def get_normalized_pattern(self, encode_settings=None):
        """Encodes"""
        normal_pattern = EmbPattern()
        transcoder = Normalizer(encode_settings)
        transcoder.transcode(self, normal_pattern)
        return normal_pattern

    def append_translation(self, x, y):
        """Appends translation to the pattern.
        All commands will be translated by the given amount,
        including absolute location commands."""
        self.add_stitch_relative(TRANSLATE, x, y, )

    def append_enable_tie_on(self, x=0, y=0):
        """Appends enable tie on.
        All starts of new stitching will be tied on"""
        self.add_stitch_relative(OPTION_ENABLE_TIE_ON, x, y)

    def append_enable_tie_off(self, x=0, y=0):
        """Appends enable tie off.
        All ends of stitching will be tied off"""
        self.add_stitch_relative(OPTION_ENABLE_TIE_OFF, x, y)

    def append_disable_tie_on(self, x=0, y=0):
        """Appends disable tie on.
        New stitching will no longer be tied on"""
        self.add_stitch_relative(OPTION_DISABLE_TIE_ON, x, y)

    def append_disable_tie_off(self, x=0, y=0):
        """Appends enable tie off.
        Ends of stitching will no longer be tied off"""
        self.add_stitch_relative(OPTION_DISABLE_TIE_OFF, x, y)



def read(f, out, settings=None):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        ctrl = byte[0]
        y = -byte[1]
        x = byte[2]
        if ctrl & 0b00100000 != 0:
            x = -x
        if ctrl & 0b01000000 != 0:
            y = -y
        ctrl &= 0b00011111
        if ctrl == 0:
            out.stitch(x, y)
            continue
        if ctrl == 0x0A:
            # Start.
            continue
        if ctrl == 0x10:
            out.move(x, y)
            continue
        if ctrl == 0x05:
            out.color_change()
            continue
        if ctrl == 0x01:
            out.trim()
            continue
        if ctrl == 0x07:
            out.end()
            return
        # We shouldn't get here.
        break

    out.end()



from .EmbConstant import *
from .WriteHelper import write_string_utf8
from .PecGraphics import get_graphic_as_string

STRIP_SEQUINS = False
MAX_JUMP_DISTANCE = 121
MAX_STITCH_DISTANCE = 121


def csv(f, values):
    string = ""
    for v in values:
        if len(string) > 0:
            string += ','
        string += ('\"%s\"' % v)
    write_string_utf8(f, string + "\n")


def write(pattern, f, settings=None):
    deltas = settings is not None and "deltas" in settings

    extends = pattern.extends()
    width = extends[2] - extends[0]
    height = extends[3] - extends[1]

    csv(f, ('#', '[VAR_NAME]', '[VAR_VALUE]'))
    count_stitches = pattern.count_stitches()
    csv(f, ('>', 'STITCH_COUNT:', str(count_stitches)))
    count_threads = pattern.count_color_changes()
    csv(f, ('>', 'THREAD_COUNT:', str(count_threads)))
    csv(f, ('>', 'EXTENTS_LEFT:', str(extends[0])))
    csv(f, ('>', 'EXTENTS_TOP:', str(extends[1])))
    csv(f, ('>', 'EXTENTS_RIGHT:', str(extends[2])))
    csv(f, ('>', 'EXTENTS_BOTTOM:', str(extends[3])))
    csv(f, ('>', 'EXTENTS_WIDTH:', str(width)))
    csv(f, ('>', 'EXTENTS_HEIGHT:', str(height)))
    write_string_utf8(f, "\n")

    if len(pattern.extras) > 0:
        csv(f, (
            '#',
            '[METADATA_NAME]',
            '[METADATA]'
        ))
        for the_key, the_value in pattern.extras.items():
            if isinstance(the_value, tuple):
                the_value = "\n" + get_graphic_as_string(the_value)
            csv(f, (
                '@',
                str(the_key),
                str(the_value)
            ))
        write_string_utf8(f, "\n")

    if len(pattern.threadlist) > 0:
        csv(f, (
            '#',
            '[THREAD_NUMBER]',
            '[HEX_COLOR]',
            '[DESCRIPTION]',
            '[BRAND]',
            '[CATALOG_NUMBER]',
            '[DETAILS]',
            '[WEIGHT]'
        ))
        for i, thread in enumerate(pattern.threadlist):
            csv(f, (
                '$',
                str(i),
                thread.hex_color(),
                thread.description,
                thread.brand,
                thread.catalog_number,
                thread.details,
                thread.weight,
            ))
        write_string_utf8(f, "\n")

    if len(pattern.stitches) > 0:
        if deltas:
            csv(f, (
                '#',
                '[STITCH_INDEX]',
                '[STITCH_TYPE]',
                '[X]',
                '[Y]',
                '[DX]',
                '[DY]'
            ))
        else:
            csv(f, (
                '#',
                '[STITCH_INDEX]',
                '[STITCH_TYPE]',
                '[X]',
                '[Y]'
            ))
        names = get_common_name_dictionary()
        current_x = 0
        current_y = 0
        for i, stitch in enumerate(pattern.stitches):
            try:
                name = names[stitch[2]]
            except IndexError:
                name = "UNKNOWN " + str(stitch[2])
            if deltas:
                csv(f, (
                    '*',
                    str(i),
                    name,
                    str(stitch[0]),
                    str(stitch[1]),
                    str(stitch[0] - current_x),
                    str(stitch[1] - current_y)
                ))
            else:
                csv(f, (
                    '*',
                    str(i),
                    name,
                    str(stitch[0]),
                    str(stitch[1]),
                ))
            current_x = stitch[0]
            current_y = stitch[1]


def get_common_name_dictionary():
    return {
        NO_COMMAND: "NO_COMMAND",
        STITCH: "STITCH",
        JUMP: "JUMP",
        TRIM: "TRIM",
        STOP: "STOP",
        END: "END",
        COLOR_CHANGE: "COLOR_CHANGE",
        SEQUIN_MODE: "SEQUIN_MODE",
        SEQUIN_EJECT: "SEQUIN_EJECT",
        SEW_TO: "SEW_TO",
        NEEDLE_AT: "NEEDLE_AT",
        SEQUENCE_BREAK: "SEQUENCE_BREAK",
        COLOR_BREAK: "COLOR_BREAK",
        TIE_ON: "TIE_ON",
        TIE_OFF: "TIE_OFF",
        FRAME_EJECT: "FRAME_EJECT",
        MATRIX_TRANSLATE: "MATRIX_TRANSLATE",
        MATRIX_SCALE: "MATRIX_SCALE",
        MATRIX_ROTATE: "MATRIX_ROTATE",
        MATRIX_RESET: "MATRIX_RESET",
        OPTION_ENABLE_TIE_ON: "OPTION_ENABLE_TIE_ON",
        OPTION_ENABLE_TIE_OFF: "OPTION_ENABLE_TIE_OFF",
        OPTION_DISABLE_TIE_ON: "OPTION_DISABLE_TIE_ON",
        OPTION_DISABLE_TIE_OFF: "OPTION_DISABLE_TIE_OFF",
        OPTION_MAX_STITCH_LENGTH: "OPTION_MAX_STITCH_LENGTH",
        OPTION_MAX_JUMP_LENGTH: "OPTION_MAX_JUMP_LENGTH",
        OPTION_IMPLICIT_TRIM: "OPTION_IMPLICIT_TRIM",
        OPTION_EXPLICIT_TRIM: "OPTION_EXPLICIT_TRIM",
        CONTINGENCY_NONE: "CONTINGENCY_NONE",
        CONTINGENCY_JUMP_NEEDLE: "CONTINGENCY_JUMP_NEEDLE",
        CONTINGENCY_SEW_TO: "CONTINGENCY_SEW_TO",
    }



from .EmbConstant import *


def read(f, out, settings=None):
    f.seek(0x80, 1)  # first block of hell if I know
    f.seek(0x80, 1)  # 2nd block of hell if I know

    while True:
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        b0 = byte[0]
        if b0 == 0xF8 or b0 == 0x87 or b0 == 91:
            break
        flags = STITCH
        command_nibble = b0 & 0xF
        if command_nibble == 0:
            flags = STITCH
        elif command_nibble == 0x1:
            flags = JUMP
        elif command_nibble == 0x7:
            flags = TRIM
        elif command_nibble >= 0x8:
            flags = COLOR_CHANGE
        dx = byte[2]
        dy = -byte[1]
        if (b0 & 0x20) != 0:
            dx = -dx
        if (b0 & 0x40) != 0:
            dy = -dy
        out.add_stitch_relative(flags, dx, dy)
    out.end()



name = "pyembroidery"

# items available at the top level (e.g. pyembroidery.read)
from .PyEmbroidery import *
from .EmbConstant import *
from .EmbPattern import EmbPattern

# items available in a sub-heirarchy (e.g. pyembroidery.PecGraphics.get_graphic_as_string)
from .PecGraphics import get_graphic_as_string



def getbit(b, pos):
    return (b >> pos) & 1


def decode_dx(b0, b1, b2):
    x = 0
    x += getbit(b2, 2) * (+81)
    x += getbit(b2, 3) * (-81)
    x += getbit(b1, 2) * (+27)
    x += getbit(b1, 3) * (-27)
    x += getbit(b0, 2) * (+9)
    x += getbit(b0, 3) * (-9)
    x += getbit(b1, 0) * (+3)
    x += getbit(b1, 1) * (-3)
    x += getbit(b0, 0) * (+1)
    x += getbit(b0, 1) * (-1)
    return x


def decode_dy(b0, b1, b2):
    y = 0
    y += getbit(b2, 5) * (+81)
    y += getbit(b2, 4) * (-81)
    y += getbit(b1, 5) * (+27)
    y += getbit(b1, 4) * (-27)
    y += getbit(b0, 5) * (+9)
    y += getbit(b0, 4) * (-9)
    y += getbit(b1, 7) * (+3)
    y += getbit(b1, 6) * (-3)
    y += getbit(b0, 7) * (+1)
    y += getbit(b0, 6) * (-1)
    return -y


def process_header_info(out, prefix, value):
    if prefix == "LA":
        out.metadata("name", value)
    elif prefix == "AU":
        out.metadata("author", value)
    elif prefix == "CP":
        out.metadata("copyright", value)
    elif prefix == "TC":
        values = [x.strip() for x in value.split(',')]
        out.add_thread({
            "hex": values[0],
            "description": value[1],
            "catalog": value[2]
        })
    else:
        out.metadata(prefix, value)


def dst_read_header(f, out):
    header = f.read(512)
    header_string = header.decode('utf8')
    for line in [x.strip() for x in header_string.split('\r')]:
        if len(line) > 3:
            process_header_info(out, line[0:2].strip(), line[3:].strip())


def dst_read_stitches(f, out):
    sequin_mode = False
    while True:
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        dx = decode_dx(byte[0], byte[1], byte[2])
        dy = decode_dy(byte[0], byte[1], byte[2])
        if byte[2] & 0b11110011 == 0b11110011:
            out.stop(dx, dy)
        elif byte[2] & 0b11000011 == 0b11000011:
            out.color_change(dx, dy)
        elif byte[2] & 0b01000011 == 0b01000011:
            out.sequin_mode(dx, dy)
            sequin_mode = not sequin_mode
        elif byte[2] & 0b10000011 == 0b10000011:
            if sequin_mode:
                out.sequin_eject(dx, dy)
            else:
                out.move(dx, dy)
        else:
            out.stitch(dx, dy)


def read(f, out, settings=None):
    dst_read_header(f, out)
    dst_read_stitches(f, out)



from .EmbConstant import *

STRIP_SEQUINS = True
FULL_JUMP = True
MAX_JUMP_DISTANCE = 127
MAX_STITCH_DISTANCE = 127


def write(pattern, f, settings=None):
    stitches = pattern.stitches
    xx = 0
    yy = 0
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2]
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy
        if data is STITCH:
            # consider bounds checking the delta_x, delta_y and raising ValueError if exceeds.
            delta_x = dx & 0xFF
            delta_y = -dy & 0xFF
            f.write(bytes(bytearray([delta_x, delta_y])))
        elif data == JUMP:
            delta_x = dx & 0xFF
            delta_y = -dy & 0xFF
            f.write(b'\x80\x04')
            f.write(bytes(bytearray([delta_x, delta_y])))
        elif data == TRIM:
            f.write(b'\x80\x80\x07\x00')
            continue
        elif data == COLOR_CHANGE:
            f.write(b'\x80\x01\x00\x00')
            continue
        elif data == STOP:
            f.write(b'\x80\x01\x00\x00')
            continue
        elif data == END:
            pass



from .EmbThreadJef import get_thread_set
from .ReadHelper import read_signed, read_int_32le, signed8


def read(f, out, settings=None):
    jef_threads = get_thread_set()
    stitch_offset = read_int_32le(f)
    f.seek(20, 1)
    count_colors = read_int_32le(f)
    f.seek(88, 1)

    for i in range(0, count_colors):
        index = abs(read_int_32le(f))
        out.add_thread(jef_threads[index % len(jef_threads)])

    f.seek(stitch_offset, 0)
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] != 0x80:
            out.stitch(signed8(b[0]), -signed8(b[1]))
            continue
        control = b[1]
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        elif control == 0x02:
            out.move(signed8(b[0]), -signed8(b[1]))
            continue
        elif control == 0x01:
            out.color_change(0, 0)
            continue
        elif control == 0x10:
            break
    out.end(0, 0)



from .EmbConstant import *
from .ReadHelper import read_int_8, signed8


# I found no copies of this file.

def read(f, out, settings=None):
    f.seek(0x200)  # Fxy with longer header.
    while True:
        stitch_type = STITCH
        b1 = signed8(read_int_8(f))
        b2 = signed8(read_int_8(f))
        command_byte = read_int_8(f)
        if command_byte is None:
            break
        if command_byte == 0x91:
            break
        if (command_byte & 0x01) == 0x01:
            stitch_type = TRIM
        if (command_byte & 0x02) == 0x02:
            stitch_type = COLOR_CHANGE
        if (command_byte & 0x20) == 0x20:
            b1 = -b1
        if (command_byte & 0x40) == 0x40:
            b2 = -b2
        out.add_stitch_relative(stitch_type, b2, b1)
    out.end()



import math

from .EmbConstant import *


class Transcoder:
    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.max_stitch = settings.get("max_stitch", float('inf'))
        self.max_jump = settings.get("max_jump", float('inf'))
        self.full_jump = settings.get("full_jump", False)
        self.strip_sequins = settings.get("strip_sequins", True)
        self.explicit_trim = settings.get("explicit_trim", True)

        self.has_tie_on = settings.get("tie_on", False)
        self.has_tie_off = settings.get("tie_off", False)
        self.long_stitch_contingency = \
            settings.get("long_stitch_contingency", CONTINGENCY_JUMP_NEEDLE)

        self.matrix = get_identity()
        translate = settings.get("translate", None)
        if translate is not None:
            try:
                m = get_translate(translate[0], translate[1])
                self.matrix = matrix_multiply(self.matrix, m)
            except IndexError:
                try:
                    m = get_translate(translate.x, translate.y)
                    self.matrix = matrix_multiply(self.matrix, m)
                except AttributeError:
                    pass
        scale = settings.get("scale", None)
        if scale is not None:
            try:
                m = get_scale(scale[0], scale[1])
                self.matrix = matrix_multiply(self.matrix, m)
            except (IndexError, TypeError):
                try:
                    m = get_scale(scale.x, scale.y)
                    self.matrix = matrix_multiply(self.matrix, m)
                except AttributeError:
                    m = get_scale(scale, scale)
                    self.matrix = matrix_multiply(self.matrix, m)
        rotate = settings.get("rotate", None)
        if rotate is not None:
            m = get_rotate(rotate)
            self.matrix = matrix_multiply(self.matrix, m)
        self.source_pattern = None
        self.destination_pattern = None
        self.position = 0
        self.color_index = -1
        self.stitch = None
        self.state_trimmed = True
        self.state_sequin_mode = False
        self.needle_x = 0
        self.needle_y = 0

    def transcode(self, source_pattern, destination_pattern):
        self.source_pattern = source_pattern
        self.destination_pattern = destination_pattern
        self.transcode_metadata()
        self.transcode_threads()
        self.transcode_stitches()
        return destination_pattern

    def transcode_metadata(self):
        """Transcodes metadata, (just moves)"""
        source = self.source_pattern.extras
        dest = self.destination_pattern.extras
        dest.update(source)

    def transcode_threads(self):
        """Transcodes threads, (just moves)"""
        source = self.source_pattern.threadlist
        dest = self.destination_pattern.threadlist
        dest.extend(source)

    def transcode_stitches(self):
        """Transcodes stitches.
        Converts middle-level commands and potentially incompatible
        commands into a format friendly low level commands."""
        source = self.source_pattern.stitches
        self.state_trimmed = True
        self.needle_x = 0
        self.needle_y = 0
        self.position = 0
        self.color_index = -1

        flags = NO_COMMAND
        for self.position, self.stitch in enumerate(source):
            p = point_in_matrix_space(self.matrix, self.stitch)
            x = p[0]
            y = p[1]
            flags = self.stitch[2]

            if flags == NO_COMMAND:
                continue

            elif flags == STITCH:
                if self.state_trimmed:
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                else:
                    self.stitch_with_contingency(x, y)
            elif flags == NEEDLE_AT:
                if self.state_trimmed:
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                else:
                    self.needle_to(x, y)
            elif flags == SEW_TO:
                if self.state_trimmed:
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                else:
                    self.sew_to(x, y)

            # Middle Level Commands.
            elif flags == FRAME_EJECT:
                self.tie_off_and_trim_if_needed()
                self.jump_to(x, y)
                self.stop_here()
            elif flags == SEQUENCE_BREAK:
                self.tie_off_and_trim_if_needed()
            elif flags == COLOR_BREAK:
                self.color_break()
            elif flags == TIE_OFF:
                self.tie_off()
            elif flags == TIE_ON:
                self.tie_on()

            # Core Commands.
            elif flags == TRIM:
                self.tie_off_and_trim_if_needed()
            elif flags == JUMP:
                self.jump_to(x, y)
            elif flags == SEQUIN_MODE:
                self.toggle_sequins()
            elif flags == SEQUIN_EJECT:
                if self.state_trimmed:
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                if not self.state_sequin_mode:
                    self.toggle_sequins()
                self.sequin_at(x, y)
            elif flags == COLOR_CHANGE:
                self.tie_off_trim_color_change()
                # If we are told to do something we do it.
                # Even if it's the first command and makes no sense.
            elif flags == STOP:
                self.stop_here()
            elif flags == END:
                self.end_here()
                break

            # On-the-fly Settings Commands.
            elif flags == OPTION_ENABLE_TIE_ON:
                self.has_tie_on = True
            elif flags == OPTION_ENABLE_TIE_OFF:
                self.has_tie_off = True
            elif flags == OPTION_DISABLE_TIE_ON:
                self.has_tie_on = False
            elif flags == OPTION_DISABLE_TIE_OFF:
                self.has_tie_off = False
            elif flags == OPTION_MAX_JUMP_LENGTH:
                x = self.stitch[0]
                self.max_jump = x
            elif flags == OPTION_MAX_STITCH_LENGTH:
                x = self.stitch[0]
                self.max_stitch = x
            elif flags == OPTION_EXPLICIT_TRIM:
                self.explicit_trim = True
            elif flags == OPTION_IMPLICIT_TRIM:
                self.explicit_trim = False
            elif flags == CONTINGENCY_NONE:
                self.long_stitch_contingency = CONTINGENCY_NONE
            elif flags == CONTINGENCY_JUMP_NEEDLE:
                self.long_stitch_contingency = CONTINGENCY_JUMP_NEEDLE
            elif flags == CONTINGENCY_SEW_TO:
                self.long_stitch_contingency = CONTINGENCY_SEW_TO
            elif flags == MATRIX_TRANSLATE:
                m = get_translate(self.stitch[0], self.stitch[1])
                self.matrix = matrix_multiply(self.matrix, m)
            elif flags == MATRIX_SCALE:
                m = get_scale(self.stitch[0], self.stitch[1])
                self.matrix = matrix_multiply(self.matrix, m)
            elif flags == MATRIX_ROTATE:
                m = get_rotate(self.stitch[0])
                self.matrix = matrix_multiply(self.matrix, m)
            elif flags == MATRIX_RESET:
                self.matrix = get_identity()
        if flags != END:
            self.end_here()

    def update_needle_position(self, x, y):
        self.needle_x = x
        self.needle_y = y

    def declare_not_trimmed(self):
        if self.state_trimmed:
            self.state_trimmed = False
            if self.color_index == -1:
                self.color_index = 0

    def add(self, flags, x=None, y=None):
        if x is None:
            x = self.needle_x
        if y is None:
            y = self.needle_y
        self.destination_pattern.stitches.append([x, y, flags])

    def lookahead_stitch(self):
        """Looks forward from current position and
         determines if anymore stitching will occur."""
        source = self.source_pattern.stitches
        for pos in range(self.position, len(source)):
            stitch = source[pos]
            flags = stitch[2]
            if flags == STITCH:
                return True
            elif flags == NEEDLE_AT:
                return True
            elif flags == SEW_TO:
                return True
            elif flags == TIE_ON:
                return True
            elif flags == SEQUIN_EJECT:
                return True
            elif flags == END:
                return False
        return False

    def color_break(self):
        """Implements color break. Should add color changes add needed only."""
        if self.color_index < 0:
            return  # We haven't stitched anything, colorbreak happens, before start. Ignore.
        if not self.state_trimmed:
            if self.has_tie_off:
                self.tie_off()
            if self.explicit_trim:
                self.trim_here()
        if not self.lookahead_stitch():
            return  # No more stitching will happen, colorchange unneeded.
        self.add(COLOR_CHANGE)
        self.color_index += 1
        self.state_trimmed = True

    def tie_off_trim_color_change(self):
        if not self.state_trimmed:
            if self.has_tie_off:
                self.tie_off()
            if self.explicit_trim:
                self.trim_here()
        self.add(COLOR_CHANGE)
        self.color_index += 1
        self.state_trimmed = True

    def tie_off_and_trim_if_needed(self):
        if not self.state_trimmed:
            self.tie_off_and_trim()

    def tie_off_and_trim(self):
        if self.has_tie_off:
            self.tie_off()
        self.trim_here()

    def tie_off(self):
        try:
            b = point_in_matrix_space(
                self.matrix,
                self.source_pattern.stitches[self.position - 1],
            )
            flags = b[2]
            if flags == STITCH or flags == NEEDLE_AT or \
                    flags == SEW_TO or flags == SEQUIN_EJECT:
                self.lock_stitch(self.needle_x, self.needle_y,
                                 b[0], b[1], self.max_stitch)
        except IndexError:
            pass  # must be an island stitch. jump-stitch-jump

    def tie_on(self):
        try:
            b = point_in_matrix_space(
                self.matrix,
                self.source_pattern.stitches[self.position + 1]
            )
            flags = b[2]
            if flags == STITCH or flags == NEEDLE_AT or \
                    flags == SEW_TO or flags == SEQUIN_EJECT:
                self.lock_stitch(self.needle_x, self.needle_y,
                                 b[0], b[1], self.max_stitch)
        except IndexError:
            pass  # must be an island stitch. jump-stitch-jump

    def trim_here(self):
        self.add(TRIM)
        self.state_trimmed = True

    def toggle_sequins(self):
        if not self.strip_sequins:
            self.add(SEQUIN_MODE)
        self.state_sequin_mode = not self.state_sequin_mode

    def jump_to_within_stitchrange(self, new_x, new_y):
        """Jumps close enough to stitch a position in x,y
        without violating the length constraints."""
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_jump
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, JUMP)
        if self.full_jump:
            if self.needle_x != new_x or self.needle_y != new_y:
                self.jump_at(new_x, new_y)
        # We are currently assuming that max_jump is also max_stitch.
        # Properly it might be the case that some format could require
        # a split constraint here where we would need to jump further
        # so that we could then stitch closer.

    def jump_to(self, new_x, new_y):
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_jump
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, JUMP)
        # above code is technically not jump_to_within_stitchrange it jumps to within jump range.
        self.jump_at(new_x, new_y)

    def jump_at(self, new_x, new_y):
        if self.state_sequin_mode:
            self.toggle_sequins()  # can't jump with sequin mode on.
        self.add(JUMP, new_x, new_y)
        self.update_needle_position(new_x, new_y)

    def stitch_with_contingency(self, new_x, new_y):
        if self.long_stitch_contingency == CONTINGENCY_SEW_TO:
            self.sew_to(new_x, new_y)
        elif self.long_stitch_contingency == CONTINGENCY_JUMP_NEEDLE:
            self.needle_to(new_x, new_y)
        else:
            self.stitch_at(new_x, new_y)

    def sew_to(self, new_x, new_y):
        """Stitches to a specific location, with the emphasis on sewing.
         Subdivides long stitches into additional stitches.
        """
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_stitch
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, STITCH)
        self.stitch_at(new_x, new_y)

    def needle_to(self, new_x, new_y):
        """Insert needle at specific location, emphasis on the needle.
        Uses jumps to avoid needle penetrations where possible.

        The limit here is the max stitch limit or jump threshold.
        If jump threshold is set low, it will insert jumps even
        between stitches it could have technically encoded values for.

        Stitches to the new location, adding jumps if needed.
        """
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_stitch
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, JUMP)
        self.stitch_at(new_x, new_y)

    def stitch_at(self, new_x, new_y):
        """Inserts a stitch at the specific location.
        Should have already been checked for constraints."""
        self.add(STITCH, new_x, new_y)
        self.update_needle_position(new_x, new_y)
        self.declare_not_trimmed()

    def sequin_at(self, new_x, new_y):
        if self.strip_sequins:
            self.add(JUMP, new_x, new_y)
        else:
            self.add(SEQUIN_EJECT, new_x, new_y)
        self.update_needle_position(new_x, new_y)
        self.declare_not_trimmed()

    def stop_here(self):
        self.add(STOP)
        self.state_trimmed = True

    def end_here(self):
        self.add(END)
        self.state_trimmed = True

    def color_change_here(self):
        self.add(COLOR_CHANGE)
        self.color_index += 1
        self.state_trimmed = True

    def position_will_exceed_constraint(self, length=None, new_x=None, new_y=None):
        """Check if the stitch is too long before trying to deal with it."""
        if length is None:
            length = self.max_stitch
        if new_x is None or new_y is None:
            p = point_in_matrix_space(self.matrix,
                                      self.stitch[0],
                                      self.stitch[1])
            new_x = p[0]
            new_y = p[1]
        distance_x = new_x - self.needle_x
        distance_y = new_y - self.needle_y
        return abs(distance_x) > length or abs(distance_y) > length

    def interpolate_gap_stitches(self, x0, y0, x1, y1, max_length, data):
        """Command sequence line to x, y, respecting length as maximum.
        This does not arrive_at, it steps to within striking distance.
        The next step can arrive at (x, y) without violating constraint.
        If these are already in range, this command will do nothing.

        returns the last stitch interpolated by the code.
        """
        transcode = self.destination_pattern.stitches
        distance_x = x1 - x0
        distance_y = y1 - y0
        if abs(distance_x) > max_length or abs(distance_y) > max_length:
            if data == JUMP and self.state_sequin_mode:
                self.toggle_sequins()  # can't jump with sequin mode on.

            # python 2,3 patch of division that could be integer.
            steps_x = math.ceil(abs(distance_x / (max_length * 1.0)))
            steps_y = math.ceil(abs(distance_y / (max_length * 1.0)))
            if steps_x > steps_y:
                steps = steps_x
            else:
                steps = steps_y
            step_size_x = distance_x / steps
            step_size_y = distance_y / steps
            qx = x0
            qy = y0
            for q in range(1, int(steps)):
                # we need the gap stitches only, not start or end stitch.
                qx += step_size_x
                qy += step_size_y
                stitch = [round(qx), round(qy), data]
                transcode.append(stitch)
                self.update_needle_position(stitch[0], stitch[1])

    def lock_stitch(self, x, y, anchor_x, anchor_y, max_length=None):
        """Tie-on, Tie-off. Lock stitch from current location towards
        anchor location.Ends again at lock location. May not exceed
        max_length in the process."""
        if max_length is None:
            max_length = self.max_stitch
        transcode = self.destination_pattern.stitches
        length = distance(x, y, anchor_x, anchor_y)
        if length > max_length:
            p = oriented(x, y, anchor_x, anchor_y, max_length)
            anchor_x = p[0]
            anchor_y = p[1]
        for amount in (.33, .66, .33, 0):
            transcode.append([
                towards(x, anchor_x, amount),
                towards(y, anchor_y, amount),
                STITCH])


def distance_squared(x0, y0, x1, y1):
    """squared of distance between x0,y0 and x1,y1"""
    dx = x1 - x0
    dy = y1 - y0
    dx *= dx
    dy *= dy
    return dx + dy


def distance(x0, y0, x1, y1):
    """distance between x0,y0 and x1,y1"""
    return math.sqrt(distance_squared(x0, y0, x1, y1))


def towards(a, b, amount):
    """amount between [0,1] -> [a,b]"""
    return (amount * (b - a)) + a


def angle_radians(x0, y0, x1, y1):
    """Angle in radians between x0,y0 and x1,y1"""
    return math.atan2(y1 - y0, x1 - x0)


def oriented(x0, y0, x1, y1, r):
    """from x0,y0 in the direction of x1,y1 in the distance of r"""
    radians = angle_radians(x0, y0, x1, y1)
    return x0 + (r * math.cos(radians)), y0 + (r * math.sin(radians))


def get_identity():
    return \
        1, 0, 0, \
        0, 1, 0, \
        0, 0, 1  # identity


def get_scale(sx, sy=None):
    if sy is None:
        sy = sx
    return \
        sx, 0, 0, \
        0, sy, 0, \
        0, 0, 1


def get_translate(tx, ty):
    return \
        1, 0, 0, \
        0, 1, 0, \
        tx, ty, 1


def get_rotate(theta):
    tau = math.pi * 2
    theta *= tau / 360
    ct = math.cos(theta)
    st = math.sin(theta)
    return \
        ct, st, 0, \
        -st, ct, 0, \
        0, 0, 1


def matrix_multiply(a, b):
    return [
        a[0] * b[0] + a[1] * b[3] + a[2] * b[6],
        a[0] * b[1] + a[1] * b[4] + a[2] * b[7],
        a[0] * b[2] + a[1] * b[5] + a[2] * b[8],
        a[3] * b[0] + a[4] * b[3] + a[5] * b[6],
        a[3] * b[1] + a[4] * b[4] + a[5] * b[7],
        a[3] * b[2] + a[4] * b[5] + a[5] * b[8],
        a[6] * b[0] + a[7] * b[3] + a[8] * b[6],
        a[6] * b[1] + a[7] * b[4] + a[8] * b[7],
        a[6] * b[2] + a[7] * b[5] + a[8] * b[8]]


def point_in_matrix_space(matrix, v0, v1=None):
    if v1 is None:
        try:
            return [
                v0[0] * matrix[0] + v0[1] * matrix[3] + 1 * matrix[6],
                v0[0] * matrix[1] + v0[1] * matrix[4] + 1 * matrix[7],
                v0[2]
            ]
        except IndexError:
            return [
                v0[0] * matrix[0] + v0[1] * matrix[3] + 1 * matrix[6],
                v0[0] * matrix[1] + v0[1] * matrix[4] + 1 * matrix[7]
                # Must not have had a 3rd element.
            ]
    return [
        v0 * matrix[0] + v1 * matrix[3] + 1 * matrix[6],
        v0 * matrix[1] + v1 * matrix[4] + 1 * matrix[7]
    ]



from .ReadHelper import read_int_8, signed8


def read(f, out, settings=None):
    f.seek(0x30, 0)
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] == 0x80:
            control = b[1]
            b = bytearray(f.read(2))
            if len(b) != 2:
                break
            if control == 0x2A:
                out.color_change()
                continue
            elif control == 0x80:
                out.move(signed8(b[0]), -signed8(b[1]))
            elif control == 0xFD:
                break
        else:
            out.stitch(signed8(b[0]), -signed8(b[1]))
    out.end()



from math import floor

blank = [
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0xF0, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F,
    0x08, 0x00, 0x00, 0x00, 0x00, 0x10,
    0x04, 0x00, 0x00, 0x00, 0x00, 0x20,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x02, 0x00, 0x00, 0x00, 0x00, 0x40,
    0x04, 0x00, 0x00, 0x00, 0x00, 0x20,
    0x08, 0x00, 0x00, 0x00, 0x00, 0x10,
    0xF0, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00
]


def get_blank():
    return [m for m in blank]


def create(width, height):
    width /= 8
    return [0x00] * width * height


def draw(points, graphic, stride=6):
    for point in points:
        try:
            try:
                graphic_mark_bit(graphic,
                                 int(point.x),
                                 int(point.y),
                                 stride)
            except AttributeError:
                graphic_mark_bit(graphic,
                                 int(point[0]),
                                 int(point[1]),
                                 stride)
        except IndexError:
            pass


def draw_scaled(extends, points, graphic, stride, buffer=5):
    if extends is None:
        draw(points, graphic, stride)
        return
    try:
        left = extends.left
        top = extends.top
        right = extends.right
        bottom = extends.bottom
    except AttributeError:
        left = extends[0]
        top = extends[1]
        right = extends[2]
        bottom = extends[3]

    diagram_width = right - left
    diagram_height = bottom - top

    graphic_width = stride * 8
    graphic_height = len(graphic) / stride

    if diagram_width == 0:
        diagram_width = 1
    if diagram_height == 0:
        diagram_height = 1

    scale_x = (graphic_width - buffer) / diagram_width
    scale_y = (graphic_height - buffer) / diagram_height

    scale = min(scale_x, scale_y)

    cx = (right + left) / 2
    cy = (bottom + top) / 2

    translate_x = -cx
    translate_y = -cy

    translate_x *= scale
    translate_y *= scale

    translate_x += graphic_width / 2
    translate_y += graphic_height / 2

    for point in points:
        try:
            try:
                graphic_mark_bit(graphic,
                                 int(floor((point.x * scale) + translate_x)),
                                 int(floor((point.y * scale) + translate_y)),
                                 stride)
            except AttributeError:
                graphic_mark_bit(graphic,
                                 int(floor((point[0] * scale) + translate_x)),
                                 int(floor((point[1] * scale) + translate_y)),
                                 stride)
        except IndexError:
            pass


def clear(graphic):
    for b in graphic:
        b = 0


def graphic_mark_bit(graphic, x, y, stride=6):
    """expressly sets the bit in the give graphic object"""
    graphic[(y * stride) + int(x / 8)] |= 1 << (x % 8)


def graphic_unmark_bit(graphic, x, y, stride=6):
    """expressly unsets the bit in the give graphic object"""
    graphic[(y * stride) + int(x / 8)] &= ~(1 << (x % 8))


def get_graphic_as_string(graphic, one="#", zero=" "):
    """Prints graphic object in text."""
    stride = 6
    if isinstance(graphic, tuple):
        stride = graphic[1]
        graphic = graphic[0]

    if isinstance(graphic, str):
        graphic = bytearray(graphic)

    list_string = [
        one if (byte >> i) & 1 else zero
        for byte in graphic
        for i in range(0, 8)
    ]
    bit_stride = 8 * stride
    bit_length = 8 * len(graphic)
    return '\n'.join(
        ''.join(list_string[m:m + bit_stride])
        for m in range(0, bit_length, bit_stride))



import math
from .ReadHelper import read_int_16be, read_int_32be, \
    read_int_8, read_string_8, signed16, signed8
from .EmbThreadShv import get_thread_set
from .EmbConstant import *


def read(f, out, settings=None):
    in_jump = False
    f.seek(0x56, 1)  # header text
    length = read_int_8(f)
    out.metadata("name", read_string_8(f, length))

    design_width = read_int_8(f)
    design_height = read_int_8(f)

    skip = math.ceil(design_height / 2.0) * design_width
    f.seek(4 + int(skip), 1)
    colors = read_int_8(f)
    f.seek(18, 1)
    threads = get_thread_set()
    stitch_per_color = {}
    for i in range(colors):
        stitch_count = read_int_32be(f)
        color_code = read_int_8(f)
        thread = threads[color_code % len(threads)]
        out.add_thread(thread)
        stitch_per_color[i] = stitch_count
        f.seek(9, 1)
    f.seek(-2, 1)

    stitches_since_stop = 0
    current_color_index = 0
    try:
        max_stitches = stitch_per_color[current_color_index]
    except IndexError:
        max_stitches = 0
    while True:
        flags = STITCH
        if in_jump:
            flags = JUMP
        b0 = read_int_8(f)
        b1 = read_int_8(f)
        if b1 is None:
            break
        if stitches_since_stop >= max_stitches:
            out.color_change()
            stitches_since_stop = 0
            current_color_index += 1
            try:
                max_stitches = stitch_per_color[current_color_index]
            except KeyError:
                max_stitches = 0xFFFFFFFF
        if b0 == 0x80:
            stitches_since_stop += 1
            if b1 == 3:
                continue
            elif b1 == 2:
                in_jump = False
                continue
            elif b1 == 1:
                stitches_since_stop += 2
                sx = signed16(read_int_16be(f))
                sy = signed16(read_int_16be(f))
                in_jump = True
                out.move(sx, sy)
                continue
        dx = signed8(b0)
        dy = signed8(b1)
        stitches_since_stop += 1
        out.add_stitch_relative(flags, dx, dy)
    out.end()



def read(f, out, settings=None):
    f.seek(0x200, 0)
    count = 0
    trimmed = False
    previous_needle = -1
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break

        y = -byte[0]
        x = byte[1]
        ctrl = byte[2]

        if ctrl & 0b01000000 != 0:
            x = -x
        if ctrl & 0b00100000 != 0:
            y = -y
        ctrl &= ~0b11100000

        if trimmed:  # any x, y gets executed regardless.
            out.move(x, y)
        else:
            out.stitch(x, y)

        if ctrl == 0:
            continue
        if ctrl == 0b00000111:  # 0x07, Initialize stitch
            continue
        if ctrl == 0b00011001:  # 0x19, start sewing again.
            trimmed = False
            continue
        trimmed = True
        if ctrl == 0b00010011:  # 0x13 trim, no color change.
            continue
        if previous_needle != ctrl:
            out.color_change()
        else:
            out.stop()
        previous_needle = ctrl



from .ReadHelper import signed8


def read_exp_stitches(f, out):
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] == 0x80:
            control = b[1]
            b = bytearray(f.read(2))  # 07 00
            if len(b) != 2:
                break
            if control == 0x80:  # Trim
                out.trim()
            elif control == 0x02:
                out.stitch(signed8(b[0]), -(signed8(b[1])))
                # This shouldn't exist.
            elif control == 0x04:  # Jump
                out.move(signed8(b[0]), -signed8(b[1]))
            elif control == 0x01:  # Colorchange
                out.color_change()
                out.move(signed8(b[0]), -signed8(b[1]))
        else:
            out.stitch(signed8(b[0]), -signed8(b[1]))
    out.end()


def read(f, out, settings=None):
    read_exp_stitches(f, out)



from .PecWriter import write_pec
from .EmbThreadPec import get_thread_set
from .WriteHelper import write_string_utf8, write_int_32le, write_int_16le, write_int_8, write_float_32le
from .EmbConstant import *

STRIP_SEQUINS = True
FULL_JUMP = True
MAX_JUMP_DISTANCE = 2047
MAX_STITCH_DISTANCE = 2047

VERSION_1 = 1
VERSION_6 = 6

PES_VERSION_1_SIGNATURE = "#PES0001"
PES_VERSION_6_SIGNATURE = "#PES0060"

EMB_ONE = "CEmbOne"
EMB_SEG = "CSewSeg"


def write(pattern, f, settings=None):
    version = VERSION_6
    truncated = False
    if settings is not None:
        version = settings.get("pes version", version)
        truncated = settings.get("truncated", truncated)
    if truncated:
        if version == VERSION_1:
            write_truncated_version_1(pattern, f)
        elif version == VERSION_6:
            write_truncated_version_6(pattern, f)
    else:
        if version == VERSION_1:
            write_version_1(pattern, f)
        elif version == VERSION_6:
            write_version_6(pattern, f)


def write_truncated_version_1(pattern, f):
    write_string_utf8(f, PES_VERSION_1_SIGNATURE)
    f.write(b'\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
    write_pec(pattern, f)


def write_truncated_version_6(pattern, f):
    chart = pattern.threadlist
    write_string_utf8(f, PES_VERSION_6_SIGNATURE)
    placeholder_pec_block = f.tell()
    write_int_32le(f, 0)  # Placeholder for PEC BLOCK
    write_pes_header_v6(pattern, f, chart, 0)
    write_int_16le(f, 0x0000)
    write_int_16le(f, 0x0000)
    current_position = f.tell()
    f.seek(placeholder_pec_block, 0)
    write_int_32le(f, current_position)
    f.seek(current_position, 0)
    # this might need that node table thing.
    write_pec(pattern, f)


def write_version_1(pattern, f):
    chart = get_thread_set()
    write_string_utf8(f, PES_VERSION_1_SIGNATURE)

    extends = pattern.extends()
    cx = (extends[2] + extends[0]) / 2.0
    cy = (extends[3] + extends[1]) / 2.0

    left = extends[0] - cx
    top = extends[1] - cy
    right = extends[2] - cx
    bottom = extends[3] - cy

    placeholder_pec_block = f.tell()
    write_int_32le(f, 0)  # Placeholder for PEC BLOCK

    if len(pattern.stitches) == 0:
        write_pes_header_v1(f, 0)
        write_int_16le(f, 0x0000)
        write_int_16le(f, 0x0000)
    else:
        write_pes_header_v1(f, 1)
        write_int_16le(f, 0xFFFF)
        write_int_16le(f, 0x0000)
        write_pes_blocks(f, pattern, chart, left, top, right, bottom, cx, cy)

    current_position = f.tell()
    f.seek(placeholder_pec_block, 0)
    write_int_32le(f, current_position)
    f.seek(current_position, 0)

    write_pec(pattern, f)


def write_version_6(pattern, f):
    chart = pattern.threadlist
    write_string_utf8(f, PES_VERSION_6_SIGNATURE)

    extends = pattern.extends()
    cx = (extends[2] + extends[0]) / 2.0
    cy = (extends[3] + extends[1]) / 2.0

    left = extends[0] - cx
    top = extends[1] - cy
    right = extends[2] - cx
    bottom = extends[3] - cy

    placeholder_pec_block = f.tell()
    write_int_32le(f, 0)  # Placeholder for PEC BLOCK

    if len(pattern.stitches) == 0:
        write_pes_header_v6(pattern, f, chart, 0)
        write_int_16le(f, 0x0000)
        write_int_16le(f, 0x0000)
    else:
        write_pes_header_v6(pattern, f, chart, 1)
        write_int_16le(f, 0xFFFF)
        write_int_16le(f, 0x0000)
        log = write_pes_blocks(f, pattern, chart, left, top, right, bottom, cx, cy)
        # In version 6 there is some node, tree, order thing.
        write_int_32le(f, 0)
        write_int_32le(f, 0)
        for i in range(0, len(log)):
            write_int_32le(f, i)
            write_int_32le(f, 0)

    current_position = f.tell()
    f.seek(placeholder_pec_block, 0)
    write_int_32le(f, current_position)
    f.seek(current_position, 0)
    write_pec(pattern, f)


def write_pes_header_v1(f, distinct_block_objects):
    write_int_16le(f, 0x01)  # scale to fit
    write_int_16le(f, 0x01)  # 0 = 100x100, 130x180 hoop
    write_int_16le(f, distinct_block_objects)


def write_pes_header_v6(pattern, f, chart, distinct_block_objects):
    write_int_16le(f, 0x01)  # 0 = 100x100, 130x180 hoop
    f.write(b'02')  # This is an 2-digit ascii number.
    write_pes_string_8(f, pattern.get_metadata("name", None))
    write_pes_string_8(f, pattern.get_metadata("category", None))
    write_pes_string_8(f, pattern.get_metadata("author", None))
    write_pes_string_8(f, pattern.get_metadata("keywords", None))
    write_pes_string_8(f, pattern.get_metadata("comments", None))
    write_int_16le(f, 0)  # OptimizeHoopChange = False
    write_int_16le(f, 0)  # Design Page Is Custom = False
    write_int_16le(f, 0x64)  # Hoop Width
    write_int_16le(f, 0x64)  # Hoop Height
    write_int_16le(f, 0)  # Use Existing Design Area = False
    write_int_16le(f, 0xC8)  # designWidth
    write_int_16le(f, 0xC8)  # designHeight
    write_int_16le(f, 0x64)  # designPageSectionWidth
    write_int_16le(f, 0x64)  # designPageSectionHeight
    write_int_16le(f, 0x64)  # p6 # 100
    write_int_16le(f, 0x07)  # designPageBackgroundColor
    write_int_16le(f, 0x13)  # designPageForegroundColor
    write_int_16le(f, 0x01)  # ShowGrid
    write_int_16le(f, 0x01)  # WithAxes
    write_int_16le(f, 0x00)  # SnapToGrid
    write_int_16le(f, 100)  # GridInterval
    write_int_16le(f, 0x01)  # p9 curves?
    write_int_16le(f, 0x00)  # OptimizeEntryExitPoints
    write_int_8(f, 0)  # fromImageStringLength
    #  String FromImageFilename
    write_float_32le(f, float(1))
    write_float_32le(f, float(0))
    write_float_32le(f, float(0))
    write_float_32le(f, float(1))
    write_float_32le(f, float(0))
    write_float_32le(f, float(0))
    write_int_16le(f, 0)  # numberOfProgrammableFillPatterns
    write_int_16le(f, 0)  # numberOfMotifPatterns
    write_int_16le(f, 0)  # featherPatternCount
    count_thread = len(chart)
    write_int_16le(f, count_thread)  # numberOfColors
    for thread in chart:
        write_pes_thread(f, thread)
    write_int_16le(f, distinct_block_objects)  # number ofdistinct blocks


def write_pes_string_8(f, string):
    if string is None:
        write_int_8(f, 0)
        return
    if len(string) > 255:
        string = string[:255]
    write_int_8(f, len(string))
    write_string_utf8(f, string)


def write_pes_string_16(f, string):
    if string is None:
        write_int_16le(f, 0)
        return
    write_int_16le(f, len(string))
    # 16 refers to the size write not the string encoding.
    write_string_utf8(f, string)


def write_pes_thread(f, thread):
    write_pes_string_8(f, thread.catalog_number)
    write_int_8(f, thread.get_red())
    write_int_8(f, thread.get_green())
    write_int_8(f, thread.get_blue())
    write_int_8(f, 0)  # unknown
    write_int_32le(f, 0xA)  # A is custom color
    write_pes_string_8(f, thread.description)
    write_pes_string_8(f, thread.brand)
    write_pes_string_8(f, thread.chart)


def write_pes_blocks(f, pattern, chart, left, top, right, bottom, cx, cy):
    if len(pattern.stitches) == 0:
        return

    write_pes_string_16(f, EMB_ONE)
    placeholder = write_pes_sewsegheader(f, left, top, right, bottom)
    write_int_16le(f, 0xFFFF)
    write_int_16le(f, 0x0000)  # FFFF0000 means more blocks exist

    write_pes_string_16(f, EMB_SEG)
    data = write_pes_embsewseg_segments(f, pattern, chart, left, bottom, cx, cy)

    sections = data[0]
    colorlog = data[1]

    current_position = f.tell()
    f.seek(placeholder, 0)
    write_int_16le(f, sections)
    f.seek(current_position, 0)  # patch final section count.

    # If there were addition embsewsegheaders or segments they would go here.

    write_int_16le(f, 0x0000)
    write_int_16le(f, 0x0000)  # 00000000 means no more blocks.

    return colorlog


def write_pes_sewsegheader(f, left, top, right, bottom):
    width = right - left
    height = bottom - top
    hoop_height = 1800
    hoop_width = 1300
    write_int_16le(f, 0)  # left
    write_int_16le(f, 0)  # top
    write_int_16le(f, 0)  # right
    write_int_16le(f, 0)  # bottom
    write_int_16le(f, 0)  # left
    write_int_16le(f, 0)  # top
    write_int_16le(f, 0)  # right
    write_int_16le(f, 0)  # bottom
    trans_x = 0
    trans_y = 0
    trans_x += float(350)
    trans_y += float(100) + height
    trans_x += hoop_width / 2
    trans_y += hoop_height / 2
    trans_x += -width / 2
    trans_y += -height / 2
    write_float_32le(f, float(1))
    write_float_32le(f, float(0))
    write_float_32le(f, float(0))
    write_float_32le(f, float(1))
    write_float_32le(f, float(trans_x))
    write_float_32le(f, float(trans_y))

    write_int_16le(f, 1)
    write_int_16le(f, 0)
    write_int_16le(f, 0)
    write_int_16le(f, int(width))
    write_int_16le(f, int(height))
    f.write(b'\x00\x00\x00\x00\x00\x00\x00\x00')

    placeholder_needs_section_data = f.tell()
    # sections
    write_int_16le(f, 0)
    return placeholder_needs_section_data


def get_as_segments_blocks(pattern, chart, adjust_x, adjust_y):
    color_index = 0
    current_thread = pattern.get_thread_or_filler(color_index)
    color_index += 1
    color_code = current_thread.find_nearest_color_index(chart)
    stitched_x = 0
    stitched_y = 0
    for command_block in pattern.get_as_command_blocks():
        block = []
        command = command_block[0][2]
        if command == JUMP:
            block.append([stitched_x - adjust_x, stitched_y - adjust_y])
            last_pos = command_block[-1]
            block.append([last_pos[0] - adjust_x, last_pos[1] - adjust_y])
            flag = 1
        elif command == COLOR_CHANGE:
            current_thread = pattern.get_thread_or_filler(color_index)
            color_index += 1
            color_code = current_thread.find_nearest_color_index(chart)
            flag = 1
            continue
        elif command == STITCH:
            for stitch in command_block:
                stitched_x = stitch[0]
                stitched_y = stitch[1]
                block.append([stitched_x - adjust_x, stitched_y - adjust_y])
            flag = 0
        else:
            continue
        yield (block, color_code, flag)


def write_pes_embsewseg_segments(f, pattern, chart, left, bottom, cx, cy):
    section = 0
    colorlog = []

    previous_color_code = -1
    flag = -1
    adjust_x = left + cx
    adjust_y = bottom + cy
    for segs in get_as_segments_blocks(pattern, chart, adjust_x, adjust_y):
        if flag != -1:
            write_int_16le(f, 0x8003)  # section end.
        segments = segs[0]
        color_code = segs[1]
        flag = segs[2]

        if previous_color_code != color_code:
            colorlog.append([section, color_code])
            previous_color_code = color_code
            # This must trigger first segment.
        write_int_16le(f, flag)
        write_int_16le(f, color_code)
        write_int_16le(f, len(segments))
        for segs in segments:
            write_int_16le(f, int(segs[0]))
            write_int_16le(f, int(segs[1]))
        section += 1

    write_int_16le(f, len(colorlog))
    for log_item in colorlog:
        write_int_16le(f, log_item[0])
        write_int_16le(f, log_item[1])

    return section, colorlog  # how many sections, how color transitions.



from .EmbConstant import *
from .ReadHelper import read_int_16le, read_int_8, signed8


# This code proved to be utterly untestable as I never found a copy of the filetype

def read(f, out, settings=None):
    f.seek(0x02, 0)
    stitches = read_int_16le(f)
    f.seek(0x100, 0)
    while True:
        b1 = read_int_8(f)
        b2 = read_int_8(f)
        b0 = read_int_8(f)
        if b0 is None:
            break
        stitch_type = STITCH
        if (b0 & 0x02) == 0:
            stitch_type = TRIM
        if b0 == 0xF8:
            break
        x = signed8(b1)
        y = -signed8(b2)
        out.add_stitch_relative(stitch_type, x, y)
    out.end()



from .EmbConstant import *
from .EmbThreadJef import get_thread_set
from .WriteHelper import write_string_utf8, write_int_32le, write_int_8, write_int_array_8

STRIP_SEQUINS = True
FULL_JUMP = True
MAX_JUMP_DISTANCE = 127
MAX_STITCH_DISTANCE = 127

# These are in mm, embroidery units are 1/10 mm
HOOP_110X110 = 0
HOOP_50X50 = 1
HOOP_140X200 = 2
HOOP_126X110 = 3
HOOP_200X200 = 4


def write(pattern, f, settings=None):
    pattern.fix_color_count()
    color_count = pattern.count_threads()
    offsets = 0x74 + (color_count * 8)
    write_int_32le(f, offsets)
    write_int_32le(f, 0x14)
    write_string_utf8(f, "20122017218088")
    write_int_8(f, 0)
    write_int_8(f, 0)
    write_int_32le(f, color_count)
    point_count = 2  # 2 bytes for END statement.
    for stitch in pattern.stitches:
        data = stitch[2]
        if data == STITCH:
            point_count += 1
        elif data == JUMP:
            point_count += 2
        elif data == COLOR_CHANGE:
            point_count += 2
        elif data == END:
            break
    write_int_32le(f, point_count)
    extends = pattern.extends()
    design_width = int(round(extends[2] - extends[0]))
    design_height = int(round(extends[3] - extends[1]))
    write_int_32le(f, get_jef_hoop_size(design_width, design_width))
    half_width = int(round(design_width / 2))
    half_height = int(round(design_height / 2))

    # distance from center of hoop.
    write_int_32le(f, half_width)
    write_int_32le(f, half_height)
    write_int_32le(f, half_width)
    write_int_32le(f, half_height)

    # distance from default 110 x 110 hoop
    x_hoop_edge = 550 - half_width
    y_hoop_edge = 550 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from default 50 x 50 hoop
    x_hoop_edge = 250 - half_width
    y_hoop_edge = 250 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from default 140 x 200 hoop
    x_hoop_edge = 700 - half_width
    y_hoop_edge = 1000 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from default 126 x 50 hoop
    x_hoop_edge = 630 - half_width
    y_hoop_edge = 550 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    jef_threads = get_thread_set()
    for thread in pattern.threadlist:
        thread_index = thread.find_nearest_color_index(jef_threads)
        write_int_32le(f, thread_index)
    for i in range(0, color_count):
        write_int_32le(f, 0x0D)

    xx = 0
    yy = 0
    for stitch in pattern.stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2]
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy
        if data == STITCH:
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == COLOR_CHANGE:
            f.write(b'\x80\x01')
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == JUMP:
            f.write(b'\x80\x02')
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == END:
            break
    f.write(b'\x80\x10')


def get_jef_hoop_size(width, height):
    if width < 500 and height < 500:
        return HOOP_50X50
    if width < 1100 and height < 1100:
        return HOOP_110X110
    if width < 1400 and height < 2000:
        return HOOP_140X200
    return HOOP_200X200


def write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge):
    if min(x_hoop_edge, y_hoop_edge) >= 0:
        write_int_32le(f, x_hoop_edge)  # left
        write_int_32le(f, y_hoop_edge)  # top
        write_int_32le(f, x_hoop_edge)  # right
        write_int_32le(f, y_hoop_edge)  # bottom
    else:
        write_int_32le(f, -1)
        write_int_32le(f, -1)
        write_int_32le(f, -1)
        write_int_32le(f, -1)



from .DstReader import dst_read_header


def b_stitch_encoding_read(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break

        ctrl = byte[0]
        y = -byte[1]
        x = byte[2]

        if ctrl & 0b01000000 != 0:
            y = -y
        if ctrl & 0b00100000 != 0:
            x = -x

        ctrl &= ~0b11100000
        if ctrl == 0:
            out.stitch(x, y)
            continue
        if ctrl & 0b00010000 != 0:
            out.end()
            return
        if ctrl & 0b00001000 != 0:
            # Set needle. Needle is: ctrl & 0b111
            if count > 1:
                out.color_change()
            continue
        if ctrl & 0b00000001 != 0:
            out.move(x, y)
    out.end()


def read(f, out, settings=None):
    dst_read_header(f, out)
    b_stitch_encoding_read(f, out)



from .DstReader import dst_read_header


def z_stitch_encoding_read(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break

        y = -byte[0]
        x = byte[1]
        ctrl = byte[2]

        if ctrl & 0b01000000 != 0:
            x = -x
        if ctrl & 0b00100000 != 0:
            y = -y

        ctrl &= ~0b11100000
        if ctrl == 0:
            out.stitch(x, y)
            continue
        if ctrl & 0b00010000 != 0:
            out.end()
            return
        if ctrl & 0b00011110 != 0:
            # Set needle = color - (ctrl >> 1)
            # No initial call to this.
            out.color_change()
            continue
        if ctrl & 0b00000001 != 0:
            out.move(x, y)
    out.end()


def read(f, out, settings=None):
    dst_read_header(f, out)
    z_stitch_encoding_read(f, out)

# def alternative_read(f,out):
#     while True:
#         stitch_type = STITCH
#         b1 = read_int_8(f)
#         b2 = read_int_8(f)
#         command_byte = read_int_8(f)
#         if command_byte is None:
#             break
#         b1 = signed8(b1)
#         b2 = signed8(b2)
#         if command_byte == 0x91:
#             break
#         if (command_byte & 0x01) == 0x01:
#             stitch_type = TRIM
#         if (command_byte & 0x02) == 0x02:
#             stitch_type = COLOR_CHANGE
#         if (command_byte & 0x20) == 0x20:
#             b1 = -b1
#         if (command_byte & 0x40) == 0x40:
#             b2 = -b2
#         out.add_stitch_relative(stitch_type, b2, b1)
#     out.end()



def find_nearest_color_index(find_color, values):
    if isinstance(find_color, EmbThread):
        find_color = find_color.color
    red = (find_color >> 16) & 0xff
    green = (find_color >> 8) & 0xff
    blue = find_color & 0xff
    closest_index = -1
    current_index = -1
    current_closest_value = float("inf")
    for t in values:
        current_index += 1
        if t is None:
            continue
        dist = color_distance_red_mean(
            red,
            green,
            blue,
            t.get_red(),
            t.get_green(),
            t.get_blue())
        if dist <= current_closest_value:  # <= choose second if they tie.
            current_closest_value = dist
            closest_index = current_index
    return closest_index


def color_distance_red_mean(
        r1, g1, b1,
        r2, g2, b2):
    red_mean = int(round((r1 + r2) / 2))
    r = int(r1 - r2)
    g = int(g1 - g2)
    b = int(b1 - b2)
    return (((512 + red_mean) * r * r) >> 8) + 4 * g * g + \
           (((767 - red_mean) * b * b) >> 8)
    # See the very good color distance paper:
    # https://www.compuphase.com/cmetric.htm


class EmbThread:

    def __init__(self):
        self.color = 0xFF000000
        self.description = None  # type: str
        self.catalog_number = None  # type: str
        self.details = None  # type: str
        self.brand = None  # type: str
        self.chart = None  # type: str
        self.weight = None  # type: str
        # description, catalog_number, details, brand, chart, weight

    def set_color(self, r, g, b):
        self.color = 0xFF000000 | (
                (r & 255) << 16) | (
                             (g & 255) << 8) | (
                             b & 255)

    def get_opaque_color(self):
        return 0xFF000000 | self.color

    def get_red(self):
        return (self.color >> 16) & 0xFF

    def get_green(self):
        return (self.color >> 8) & 0xFF

    def get_blue(self):
        return self.color & 0xFF

    def find_nearest_color_index(self, values):
        return find_nearest_color_index(self.color, values)

    def hex_color(self):
        return "#%02x%02x%02x" % (
            self.get_red(), self.get_green(), self.get_blue())

    def set_hex_color(self, hex_string):
        h = hex_string.lstrip('#')
        size = len(h)
        if size == 6 or size == 8:
            self.color = int(h[:6], 16)
        elif size == 4 or size == 3:
            self.color = int(h[2] + h[2] + h[1] + h[1] + h[0] + h[0], 16)



NO_COMMAND = -1
STITCH = 0
JUMP = 1
TRIM = 2
STOP = 3
END = 4
COLOR_CHANGE = 5
SEQUIN_MODE = 6
SEQUIN_EJECT = 7

# Stitch with implied contingency.
SEW_TO = 0xB0
NEEDLE_AT = 0xB1

SEQUENCE_BREAK = 0xE1
COLOR_BREAK = 0xE2
TIE_ON = 0xE4
TIE_OFF = 0xE5
FRAME_EJECT = 0xE9

MATRIX_TRANSLATE = 0xC0
MATRIX_SCALE = 0xC1
MATRIX_ROTATE = 0xC2
MATRIX_RESET = 0xC3

OPTION_ENABLE_TIE_ON = 0xD1
OPTION_ENABLE_TIE_OFF = 0xD2
OPTION_DISABLE_TIE_ON = 0xD3
OPTION_DISABLE_TIE_OFF = 0xD4
OPTION_MAX_STITCH_LENGTH = 0xD5
OPTION_MAX_JUMP_LENGTH = 0xD6
OPTION_EXPLICIT_TRIM = 0xD7
OPTION_IMPLICIT_TRIM = 0xD8

CONTINGENCY_NONE = 0xF0
CONTINGENCY_JUMP_NEEDLE = 0xF1
CONTINGENCY_SEW_TO = 0xF2

# Eventually the commands are supposed to be limited to 255 thereby
# allowing additional information like, color change to color in position n
# To be stored in the higher level bits.
COMMAND_MASK = 0xFF



from xml.etree.cElementTree import Element, ElementTree, SubElement


STRIP_SEQUINS = True
# technically I could use svg to draw a sequin as a 2 element circle path.
FULL_JUMP = False  # just drops all jumps anyway.
MAX_JUMP_DISTANCE = float('inf')
MAX_STITCH_DISTANCE = float('inf')

NAME_SVG = "svg"
ATTR_VERSION = "version"
VALUE_SVG_VERSION = "1.1"
ATTR_XMLNS = "xmlns"
VALUE_XMLNS = "http://www.w3.org/2000/svg"
ATTR_XMLNS_LINK = "xmlns:xlink"
VALUE_XLINK = "http://www.w3.org/1999/xlink"
ATTR_XMLNS_EV = "xmlns:ev"
VALUE_XMLNS_EV = "http://www.w3.org/2001/xml-events"
ATTR_WIDTH = "width"
ATTR_HEIGHT = "height"
ATTR_VIEWBOX = "viewBox"
NAME_PATH = "path"
ATTR_DATA = "d"
ATTR_FILL = "fill"
ATTR_STROKE = "stroke"
ATTR_STROKE_WIDTH = "stroke-width"
VALUE_NONE = "none"


def write(pattern, f, settings=None):
    """Writes an svg file of the stitchblocks."""

    root = Element(NAME_SVG)
    root.set(ATTR_VERSION, VALUE_SVG_VERSION)
    root.set(ATTR_XMLNS, VALUE_XMLNS)
    root.set(ATTR_XMLNS_LINK, VALUE_XLINK)
    root.set(ATTR_XMLNS_EV, VALUE_XMLNS_EV)
    extends = pattern.extends()
    width = extends[2] - extends[0]
    height = extends[3] - extends[1]
    root.set(ATTR_WIDTH, str(width))
    root.set(ATTR_HEIGHT, str(height))
    viewbox = \
        str(extends[0]) + " " +\
        str(extends[1]) + " " +\
        str(width) + " " +\
        str(height)
    root.set(ATTR_VIEWBOX, viewbox)

    for stitchblock in pattern.get_as_stitchblock():
        block = stitchblock[0]
        thread = stitchblock[1]
        path = SubElement(root, NAME_PATH)
        data = "M"
        for stitch in block:
            x = stitch[0]
            y = stitch[1]
            data += " " + str(x) + "," + str(y)
        path.set(ATTR_DATA, data)
        path.set(ATTR_FILL, VALUE_NONE)
        path.set(ATTR_STROKE, thread.hex_color())
        path.set(ATTR_STROKE_WIDTH, "3")
    tree = ElementTree(root)
    tree.write(f)



from .EmbConstant import *
from .ReadHelper import read_int_8, signed8


def read(f, out, settings=None):
    f.seek(0x2000, 0)
    while True:
        stitch_type = STITCH

        x = read_int_8(f)
        y = read_int_8(f)
        command_byte = read_int_8(f)
        if command_byte is None:
            break
        x = signed8(x)
        y = -signed8(y)
        if (command_byte & 0x20) == 0x20:
            y = -y
        if (command_byte & 0x40) == 0x40:
            x = -x
        if (command_byte & 0x01) == 0x01:
            stitch_type = COLOR_CHANGE
        if (command_byte & 0x02) == 0x02:
            stitch_type = TRIM
        out.add_stitch_relative(stitch_type, x, y)
    out.end()
